/*
Programm for digital humidity meter with SHT11 sensor.

Version 1.0 - January 2003

(c) OK2XDX

*/


#include "Reg51.h"
#include <math.h>


enum {TEMP,HUMI};

sbit DDATA = P1 ^ 0;				// pins where SHT11 is connected
sbit SCK = P1 ^ 1;
sbit Calibration = P2 ^ 5;
sbit Up = P2 ^ 7;
sbit Down = P2 ^ 6;
sbit No_press = P2 ^ 4;
unsigned int N,i;

float Temperature, Humidity, DP, Tlak;
unsigned char checksum, buf[8];
unsigned char Err=0;
unsigned char CRC = 0;
unsigned char error = 0;

int Rozdil;

unsigned char code Popis [][8] = {{"Pressure"},{" calib. "},{"   OK   "},{"Sensor  "},{"error!  "}};

unsigned char code CRC_Table [] ={0, 49, 98, 83, 196, 245,166, 151,185, 136,219, 234,125, 76, 31, 46, 67, 114,33, 16,
135, 182,229, 212,250, 203,152, 169,62, 15, 92, 109,134, 183,228, 213,66, 115,32, 17,
63, 14, 93, 108,251, 202,153, 168,197, 244,167, 150,1, 48, 99, 82, 124, 77, 30, 47,
184, 137,218, 235,61, 12, 95, 110,249, 200,155, 170,132, 181,230, 215,64, 113,34, 19,
126, 79, 28, 45, 186, 139,216, 233,199, 246,165, 148,3, 50, 97, 80, 187, 138,217, 232,
127, 78, 29, 44, 2, 51, 96, 81, 198, 247,164, 149,248, 201,154, 171,60, 13, 94, 111,
65, 112,35, 18, 133, 180,231, 214,122, 75, 24, 41, 190, 143,220, 237,195, 242,161, 144,
7, 54, 101, 84, 57, 8, 91, 106,253, 204,159, 174,128, 177,226, 211,68, 117,38, 23,
252, 205,158, 175,56, 9, 90, 107,69, 116,39, 22, 129, 176,227, 210,191, 142,221, 236,
123, 74, 25, 40, 6, 55, 100, 85, 194, 243,160, 145,71, 118,37, 20, 131, 178,225, 208,
254, 207,156, 173,58, 11, 88, 105,4, 53, 102, 87, 192, 241,162, 147,189, 140,223, 238,
121, 72, 27, 42, 193, 240,163, 146,5, 52, 103, 86, 120, 73, 26, 43, 188, 141,222, 239,
130, 179,224, 209,70, 119,36, 21, 59, 10, 89, 104,255, 206,157, 172};

#define noACK 0
#define ACK 1
#define STATUS_REG_W 0x06 //000 0011 0
#define STATUS_REG_R 0x07 //000 0011 1
#define MEASURE_TEMP 0x03 //000 0001 1
#define MEASURE_HUMI 0x05 //000 0010 1
#define RESET 0x1e //000 1111 0

extern unsigned char CR(unsigned char A, unsigned char B);				// XOR written in ASM

//Variable of LCD control

extern void LCD_ini( void ); 
extern void LCDA( void ); 											// sets cursor to given part of the display (4x8)
extern void LCDB( void ); 
extern void LCDC( void ); 
extern void LCDD( void ); 
extern void LCD_write( unsigned char LCD_byte); 

//Variable of I2C

extern void IISTART (void);
extern void IIWR (unsigned char LCD_byte);
extern unsigned char IIRE(bit NOACK); 
extern void IISTOP(void);


/*-------------------------------------------------------------------------------
Priprava pameti pro cteni/zapis
---------------------------------------------------------------------------------*/

void Priprava (unsigned char	 Adresa){

	IISTART ();													// Start condition
	IIWR (0xA0);				// pise adresu pameti, A2,1,0 a posledni je 0=write
	IIWR (Adresa);								// lower part of address

}

/*-------------------------------------------------------------------------------
Waits for button release
---------------------------------------------------------------------------------*/

void Uvolni_tlac(void) {
	while (Up == 0 | Down == 0) {;}
}

/*-------------------------------------------------------------------------------
Zmeri Xkrat hodnotu na danem kanale
---------------------------------------------------------------------------------*/

unsigned int Zmer_AD(unsigned int Mereni)
{
unsigned int i;
unsigned long int Suma;

	Suma = 0;
	for (i = 0; i < Mereni; i++){
		Suma = Suma + ADS7822();
	}
	Suma = Suma / Mereni;
	return (Suma);
}


/*-------------------------------------------------------------------------------
Calculates CRC
---------------------------------------------------------------------------------*/

void Calc_CRC(unsigned char X)
{
unsigned char N;
	N = (X ^ CRC);							//XRL
	CRC = CRC_Table[N];
}


/*-------------------------------------------------------------------------------
Recalculates FLOAT to ASCII and sends to display and serial port. Shorter then PRINTF
---------------------------------------------------------------------------------*/

void Prevod_ASCII(float Hodnota, unsigned char Znak) 
{

int IntHodnota;
unsigned char Zbytek,i,Shift;

	if (Znak != 'P')
	{   
		IntHodnota = (int)(10*Hodnota);
		Shift = 0;
	}
	else
	{
		IntHodnota = (int)(Hodnota);
		Shift = 1;
	}


	if (IntHodnota < 0) {
		buf [0] = 45;								// sign -
		IntHodnota = abs (IntHodnota);
		}
	else                                             
		buf [0] = 32;

	for (i = 5+Shift; i > 0; i--) {					// recalculate float and store to BUF
		Zbytek = IntHodnota % 10;
		IntHodnota = (IntHodnota/10);
		buf [i] = Zbytek + 48;					//  ASCII shift
	}

     if (buf[1] == 48){							// remove zeros
   	buf[1] = 32;
   	if (buf[2] == 48) {
   		buf[2] = 32;
   		if (buf[0] == 45) {
   			buf[0] = 32;
   			buf[2] = 45;						// shift -
   		}
   	}
   }
   if ((buf[1] == 32) & (buf[0] == 45)){
   	buf[0] = 32;
   	buf[1] = 45;
   }													

	if (Znak != 'P')
	{   
		buf[6] = buf[5];							//adding '.'
		buf[5] = 46;
		buf[7] = ' ';
	}
	else
	{
//		buf[6] = ' ';
		buf[7] = ' ';
	}
	buf[0] = Znak;
	buf[1] = ' ';
	buf[0] = ' ';

/*
	if (Znak == 'T')							// display unit C or %
	{
		buf[7] = 'C';
	}
	if (Znak == 'V')
	{
		buf[7] = '%';
	}

	if (Znak == 'D')
	{
		buf[7] = '%';
	}
	if (Znak == 'P')
	{
		buf[7] = 'h';
	}
*/

// send
	
	for (i = 0; i < 8; i++) {
	SBUF = buf[i];				// sends to RS232
	LCD_write (buf[i]);		// sends to the display
	while (TI == 0) {;}		// wait for sending
	TI = 0;
	}
	SBUF = ';';				// sends to RS232 ;
	while (TI == 0) {;}		// wait for sending
	TI = 0;

}


//----------------------------------------------------------------------------------
char s_write_byte(unsigned char Xvalue)
//----------------------------------------------------------------------------------
// writes a byte on the Sensibus and checks the acknowledge
{
unsigned char i,error=0;

	for (i=0x80;i>0;i/=2) 		//shift bit for masking
	{ if (i & Xvalue) DDATA=1; //masking value with i , write to SENSI-BUS
		else DDATA=0;
		SCK=1; 						//clk for SENSI-BUS
		_nop_();_nop_();_nop_(); //pulswith approx. 5 us
		_nop_();_nop_();_nop_();
		SCK=0;
		_nop_();_nop_();_nop_();
	}

	DDATA=1; //release DDATA-line
	SCK=1; //clk #9 for ack
	_nop_();_nop_();_nop_();
	error=DDATA; //check ack (DDATA will be pulled down by SHTXX)
	SCK=0;
	return error; //error=1 in case of no acknowledge
}

//----------------------------------------------------------------------------------
char s_read_byte(unsigned char ack, unsigned char Co)
//----------------------------------------------------------------------------------
// reads a byte form the Sensibus and gives an acknowledge in case of "ack=1"
{

unsigned char val=0;
unsigned int i=0;

	DDATA=1; //release DDATA-line

	if (Co == 0)
	{
		for (i=0x80;i>0;i/=2) //shift bit for masking
		{ SCK=1; //clk for SENSI-BUS
			_nop_();_nop_();_nop_(); //pulswith approx. 5 us
			_nop_();_nop_();_nop_();
			if (DDATA) val=(val | i); //read bit
			SCK=0;
			_nop_();_nop_();_nop_(); //pulswith approx. 5 us
			_nop_();_nop_();_nop_();
		}
	}

	if (Co == 1)				// CRC has to be read in reversed order
	{
		for (i=0x01;i<255;i*=2) //shift bit for masking
		{ SCK=1; //clk for SENSI-BUS
			_nop_();_nop_();_nop_(); //pulswith approx. 5 us
			_nop_();_nop_();_nop_();
			if (DDATA) val=(val | i); //read bit
			SCK=0;
			_nop_();_nop_();_nop_(); //pulswith approx. 5 us
			_nop_();_nop_();_nop_();
		}
	}

	DDATA=!ack; //in case of "ack==1" pull down DDATA-Line
	SCK=1; //clk #9 for ack
	_nop_();_nop_();_nop_(); //pulswith approx. 5 us
	_nop_();_nop_();_nop_();
	SCK=0;
	DDATA=1; //release DDATA-line
	_nop_();_nop_();_nop_();
return val;
}

//----------------------------------------------------------------------------------
void s_transstart(void)
//----------------------------------------------------------------------------------
// generates a transmission start
//       _____         ________
// DDATA:     |_______|
//           ___     ___
// SCK : ___|   |___|   |______
{
	DDATA=1; SCK=0; //Initial state
	_nop_();
	SCK=1;
	_nop_();
	DDATA=0;
	_nop_();
	SCK=0;
	_nop_();_nop_();_nop_();
	_nop_();_nop_();_nop_();
	SCK=1;
	_nop_();
	DDATA=1;
	_nop_();
	SCK=0;
}

//----------------------------------------------------------------------------------
void s_connectionreset(void)
//----------------------------------------------------------------------------------
// communication reset: DDATA-line=1 and at least 9 SCK cycles followed by transstart
//       _____________________________________________________         ________
// DDATA: 																	  |_______|
// 			_    _    _    _    _    _    _    _    _        ___     ___
// SCK : __| |__| |__| |__| |__| |__| |__| |__| |__| |______|   |___|   |______
{
unsigned char i;
DDATA=1; SCK=0; //Initial state
	for(i=0;i<9;i++) //9 SCK cycles
	{ SCK=1;
		_nop_();_nop_();_nop_();
		SCK=0;
		_nop_();_nop_();_nop_();
	}
	s_transstart(); //transmission start
}

//----------------------------------------------------------------------------------
char s_read_statusreg(void)
//----------------------------------------------------------------------------------
// reads the status register with checksum (8-bit)
{
unsigned char error=0;
unsigned char Hodn;
	s_transstart(); //transmission start
	error=s_write_byte(STATUS_REG_R); //send command to sensor
	Hodn =s_read_byte(ACK,0); //read status register (8-bit)
	checksum=s_read_byte(noACK,1); //read checksum (8-bit)

return Hodn; //error=1 in case of no response form the sensor
}

/*
//----------------------------------------------------------------------------------
char s_write_statusreg(unsigned char *p_value)
//----------------------------------------------------------------------------------
// writes the status register with checksum (8-bit)
{
//unsigned char error=0;
s_transstart(); //transmission start
error+=s_write_byte(STATUS_REG_W);//send command to sensor
error+=s_write_byte(*p_value); //send value of status register
return error; //error>=1 in case of no response form the sensor
}

*/
//----------------------------------------------------------------------------------
int s_measure(unsigned char mode)
//----------------------------------------------------------------------------------
// makes a measurement (humidity/temperature) with checksum
{
//unsigned error=0;
unsigned int i;
int Vysledek;
unsigned char Cteny_bajt;

	CRC = 0;				//inicializace
	s_transstart(); //transmission start
	switch(mode){ //send command to sensor
	case TEMP : error+=s_write_byte(MEASURE_TEMP); Calc_CRC(MEASURE_TEMP); break;
	case HUMI : error+=s_write_byte(MEASURE_HUMI); Calc_CRC(MEASURE_HUMI); break;
	default : break;
	}

	for (i=0;i<65535;i++) if(DDATA==0) break; //wait until sensor has finished the measurement
	if(DDATA) error+=1; // or timeout (~2 sec.) is reached
	Cteny_bajt = (s_read_byte(ACK,0) & 0xFF);
	Vysledek = Cteny_bajt;
	Calc_CRC(Cteny_bajt);				//prepocita CRC
	Cteny_bajt = (s_read_byte(ACK,0) & 0xFF);	
	Calc_CRC(Cteny_bajt);				//prepocita CRC
	Vysledek = ((Vysledek << 8) + Cteny_bajt); //read the second byte (LSB)
	checksum =s_read_byte(noACK,1); //read checksum

return Vysledek;
}

//----------------------------------------------------------------------------------
void init_uart()       
//----------------------------------------------------------------------------------
//9600 bps @ 11.059 MHz
{
	SCON = 0x52;
	TMOD = 0x20;
	TCON = 0x69;
	TH1 = 0xfd;
}


//----------------------------------------------------------------------------------------
void calc_sth11(void)
//----------------------------------------------------------------------------------------
// calculates temperature [  C] and humidity [%RH]
// input : humi [Ticks] (12 bit)
// temp [Ticks] (14 bit)
// output: humi [%RH]
// temp [  C]
{ 
float rh_lin; // rh_lin: Humidity linear
float rh_true; // rh_true: Temperature compensated humidity
float t_C; // t_C : Temperature [  C]

t_C = Temperature*0.01 - 40; //calc. temperature from ticks to [  C]
rh_lin=-0.0000028*Humidity*Humidity + 0.0405*Humidity - 4.0; //calc. humidity from ticks to [%RH]
rh_true=(t_C-25)*(0.01+0.00008*Humidity)+rh_lin; //calc. temperature compensated humidity [%RH]

Temperature = t_C;
Humidity = rh_true;

}

//--------------------------------------------------------------------
void Write_error(void)
//--------------------------------------------------------------------
{ 
unsigned char i;

	LCDA();
	for (i = 0; i < 8; i++) {
		LCD_write (Popis[3][i]);		// sends to the display
	}
	LCDB();
	for (i = 0; i < 8; i++) {
		LCD_write (Popis[4][i]);		// sends to the display
	}


}


//--------------------------------------------------------------------
void calc_dewpoint(float h,float t)
//--------------------------------------------------------------------
// calculates dew point
// input: humidity [%RH], temperature [  C]
// output: dew point [  C]
{ float logEx;
	logEx=0.66077+7.5*t/(237.3+t)+(log10(h)-2);
//	DP = ((0.66077 - logEx)*237.3)/(logEx - 8.16077);
	DP = (logEx - 0.66077)*237.3/(0.66077+7.5-logEx);
}


//----------------------------------------------------------------------------------

void Make_calib(void)
{
unsigned char i;
unsigned int Tlak_cal;

	LCDA();
	for (i = 0; i < 8; i++) {
		LCD_write (Popis[0][i]);		// sends to the display
	}
	LCDB();
	for (i = 0; i < 8; i++) {
		LCD_write (Popis[1][i]);		// sends to the display
	}

	Tlak_cal = 1013;
	LCDC();
	Prevod_ASCII(Tlak_cal,'P');

	while (~Calibration)
	{
		if (~Up)
		{
			Tlak_cal++;
			Uvolni_tlac();		// ceka na pusteni tlacitka
			LCDC();
			Prevod_ASCII(Tlak_cal,'P');
		}
		if (~Down)
		{
			Tlak_cal--;
			Uvolni_tlac();		// ceka na pusteni tlacitka
			LCDC();
			Prevod_ASCII(Tlak_cal,'P');
		}
	}
	LCDC();
	for (i = 0; i < 8; i++) {
		LCD_write (Popis[2][i]);		// sends to the display
	}

	Tlak = Zmer_AD(256);
	Tlak = ((Tlak + 389.12) / 36.384) * 10;
	Rozdil = Tlak_cal - (int)Tlak;

	Priprava(1);			// priprava zapisu do pameti
	IIWR(Rozdil >> 8);	
	IIWR(Rozdil);			// zapise hodnotu HI, pak LO
	IISTOP();
	
	while(1);


}

void main()

{
	init_uart();
	LCD_ini();	
	s_connectionreset();			// make softstart

	if (~Calibration) Make_calib();

	Priprava(1);			// priprava cteni z pameti
	IISTART();
	IIWR(0xA1);				// cteni
	Rozdil = IIRE(0);
	Rozdil = (Rozdil << 8) + IIRE(1);
	IISTOP();

while(1)
	{
		Calc_CRC(MEASURE_HUMI);
		Calc_CRC(0x09);
		Calc_CRC(0x31);

		s_connectionreset();
 		Humidity = s_measure(HUMI); //measure humidity
		if (CRC != checksum)
		{
			Err = 1;
			Write_error();
			s_connectionreset();			// make softstart
      }
		else Err = 0;
		Temperature = s_measure(TEMP); //measure temperature
		if (CRC != checksum)
		{
			Err = 1;
			Write_error();
			s_connectionreset();			// make softstart
		}
		else Err = 0;
		
		calc_sth11(); //calculate humidity, temperature
		calc_dewpoint(Humidity,Temperature);

		Tlak = Zmer_AD(256);

		Tlak = (((Tlak + 389.12) / 36.384) * 10) + Rozdil;
		
		LCDA();						// sets segment A at LCD
		Prevod_ASCII(Temperature,'T');		// writes value

		LCDB();
		Prevod_ASCII(Humidity,'H');

		LCDC();
		Prevod_ASCII(DP,'D');

		if (No_press)
		{
			LCDD();
			Prevod_ASCII(Tlak,'P');
		}

		SBUF = 13;					// send <CR>
		while (TI == 0) {;}		// wait for sending
		TI = 0;

		for (N = 0; N < 170; N++)                  //waits 1sec
		{
			for (i = 0; i < 500; i++);
		}

		
	}
}

